/***************************************************************************
 *   Copyright (C) 2004 by Stefan Baur                                     *
 *   stefan@c0n.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <GL/gl.h>
#include "glwater.h"
#include "string.h"

	//We define this for now... as they are experimantal at the moment
#define NO_GL_EXTENSIONS

#ifndef NO_GL_EXTENSIONS
extern "C" void (*glXGetProcAddressARB(const char *procName))(void);
#endif

#include <sys/time.h>
#include <math.h>

GLWater::GLWater()
	{
	currentTime=0;
	lastWaterStepTime=0;

	Settings.bgTexID=0;

	vbo_vtx=0;
	vbo_idx=0;
	vbo_tc=0;

	water_data=0;
	water_vtx=0;
	water_idx=0;
	water_tc=0;
	water_diff=0;
	}

GLWater::~GLWater()
	{
	quit();
	}

void GLWater::quit()
	{
	if(water_data)delete[] water_data;water_data=0;
	if(water_vtx)delete[] water_vtx;water_vtx=0;
	if(water_tc)delete[] water_tc;water_tc=0;
	if(water_diff)delete[] water_diff;water_diff=0;
	if(water_idx)delete[] water_idx;water_idx=0;
	if(Settings.bgTexID!=0 && glIsTexture(Settings.bgTexID))glDeleteTextures(1,&Settings.bgTexID);
	Settings.bgTexID=0;
	}

bool GLWater::isExtensionSupported(char *str)
	{
#ifdef NO_GL_EXTENSIONS
	str=str;//do nothing and keep compiler happy...
#else
	char *extensions=(char*)glGetString(GL_EXTENSIONS);
	if(glXGetProcAddressARB==0)return false;
	if(strstr(extensions,str))return true;
#endif
	return false;
	}

void GLWater::init()
	{
	int x,y;
	float vtx_stepx,vtx_stepy,vtx_offsx,vtx_offsy,vtx_x,vtx_y;
	INDEX *idx;
	VERTEX *vtx;

	glDisable(GL_DEPTH_TEST);
	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glDisable(GL_BLEND);

	glShadeModel( GL_SMOOTH ); //hey OpenGL, we want some nice rendering :)
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glHint( GL_POINT_SMOOTH_HINT, GL_NICEST );


		//extensions
	if((OpenGLExtensions.ext_VBO=isExtensionSupported((char*)"GL_ARB_vertex_buffer_object")))
		{
#ifndef NO_GL_EXTENSIONS
		ext.glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) glXGetProcAddressARB("glGenBuffersARB");
		ext.glBindBufferARB = (PFNGLBINDBUFFERARBPROC) glXGetProcAddressARB("glBindBufferARB");
		ext.glBufferDataARB = (PFNGLBUFFERDATAARBPROC) glXGetProcAddressARB("glBufferDataARB");
		ext.glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) glXGetProcAddressARB("glDeleteBuffersARB");
#endif
		}
	else
		{
		ext.glGenBuffersARB = 0;
		ext.glBindBufferARB = 0;
		ext.glBufferDataARB = 0;
		ext.glDeleteBuffersARB = 0;
		}


		//water effects date storage
	water_data=new WATERTYPE[Settings.water_res_x*Settings.water_res_y*2];
	water_vtx=new VERTEX[Settings.water_res_x*Settings.water_res_y];
	water_tc=new TEXCOORD[Settings.water_res_x*Settings.water_res_y];
	water_diff=new TEXCOORD[Settings.water_res_x*Settings.water_res_y];
	water_idx=new INDEX[(Settings.water_res_x-1)*(Settings.water_res_y-1)*4];


		//initialize with zero
	for(y=0;y<Settings.water_res_y;y++)
		{
		for(x=0;x<Settings.water_res_x;x++)
			{
			water_data[y*Settings.water_res_x+x]=0;
			water_data[y*Settings.water_res_x+x + Settings.water_res_x*Settings.water_res_y]=0;

			water_vtx[y*Settings.water_res_x+x].x=0;water_vtx[y*Settings.water_res_x+x].y=0;
			water_tc[y*Settings.water_res_x+x].u=0;water_tc[y*Settings.water_res_x+x].v=0;
			water_diff[y*Settings.water_res_x+x].u=0;water_diff[y*Settings.water_res_x+x].v=0;
			}
		}


	idx=water_idx;

		//Create Indices once
	for(y=0;y<Settings.water_res_y-1;y++)
		{
		for(x=0;x<Settings.water_res_x-1;x++)
			{
				// (0,0)
			*idx=y*Settings.water_res_x+x; idx++;
				// (0,1)
			*idx=(y+1)*Settings.water_res_x+x; idx++;
				// (1,1)
			*idx=(y+1)*Settings.water_res_x+x+1; idx++;
				// (1,0)
			*idx=y*Settings.water_res_x+x+1; idx++;
			}
		}

		//Create Plain TexCoords
	for(y=0;y<Settings.water_res_y;y++)
		{
		for(x=0;x<Settings.water_res_x;x++)
			{
			water_tc[y*Settings.water_res_x+x].u=(float)x/((float)Settings.water_res_y-1);
			water_tc[y*Settings.water_res_x+x].v=(float)y/((float)Settings.water_res_y-1);
			}
		}

		//Create Vertices once
	vtx_offsx=-1;
	vtx_offsy=-1;
	vtx_stepx=2.0f/((float)Settings.water_res_x-1);
	vtx_stepy=2.0f/((float)Settings.water_res_y-1);
	vtx=water_vtx;

	for(y=0;y<Settings.water_res_y;y++)
		{
		vtx_y=((float)y)*vtx_stepy+vtx_offsy;

		for(x=0;x<Settings.water_res_x;x++)
			{
			vtx_x=((float)x)*vtx_stepx+vtx_offsx;
			vtx->x=vtx_x;vtx->y=vtx_y;vtx++;
			}
		}

		//build vertex buffers
	if(OpenGLExtensions.ext_VBO)
		{
		ext.glGenBuffersARB(1,&vbo_idx);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_idx );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, (Settings.water_res_x-1)*(Settings.water_res_y-1)*4*sizeof(INDEX), water_idx, GL_STATIC_DRAW_ARB);

		ext.glGenBuffersARB(1,&vbo_vtx);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_vtx );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, Settings.water_res_x*Settings.water_res_y*sizeof(VERTEX), water_vtx, GL_STATIC_DRAW_ARB);

		ext.glGenBuffersARB(1,&vbo_tc);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_tc );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, Settings.water_res_x*Settings.water_res_y*sizeof(TEXCOORD), water_tc, GL_STATIC_DRAW_ARB); //GL_DYNAMIC_DRAW_ARB); //FIXME updateable?
		}
	}

bool GLWater::update() //TODO: transparently return the ms, our timer should wait next time
	{
	int x,y;
	float tc_stepx,tc_stepy,tc_u,tc_v,vtx_x,vtx_y;
	TEXCOORD *tc,*tc_diff;
	WATERTYPE *w_current;


	if(Settings.realtime) //realtime ...
		{
		double currentTick;
		timeval tv;

			//calculate time difference since last update()
		gettimeofday(&tv,0);
		currentTick = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;

		diffTime = currentTick - lastTick;
		lastTick = currentTick;
		diffTime = (diffTime > 0.5) ? 0.5 : (diffTime < 0.005) ? 0.005 : diffTime;
		}
	else	//always do 1 update per frame
		{
		diffTime = Settings.water_step_time;
		}

	currentTime+=diffTime;

	if(!(lastWaterStepTime<currentTime))
		return false; //no update needed since nothing has to change yet.


		//calculate each step (without visual update)
	while(lastWaterStepTime<currentTime)
		{
		calcWaterStep();
		lastWaterStepTime+=Settings.water_step_time;

		w_current=water_data+(Settings.water_res_x*Settings.water_res_y)*((waterStep+1)&1) +Settings.water_res_x;


		//TODO: Add effects, make them configurable and randomized

		/*if((waterStep%340)==0) //center blob
			w_current[Settings.water_res_x*Settings.water_res_y/2+Settings.water_res_x/2]=44340;*/

			//big psycedelic center blob
		if(Settings.effect_bigblob)
			{
			w_current[Settings.water_res_x*Settings.water_res_y/2+Settings.water_res_x/2]=(int)(10000*10*sin(lastWaterStepTime*10)*sin(lastWaterStepTime/20)*sin(lastWaterStepTime/20));
			}

			//circling blob
		if(Settings.effect_whirl)
			{
			vtx_x=0.8f*sin(lastWaterStepTime)/2+0.5;
			vtx_y=0.8f*cos(lastWaterStepTime)/2+0.5;
			w_current[ (int)( vtx_y*Settings.water_res_y )*Settings.water_res_x+(int)( vtx_x*Settings.water_res_x )]=(int)(10000*10*sin(lastWaterStepTime/3.1415f));
			}

		if(Settings.effect_rain)
			{
			static int rnd;

			rnd+=(int)(Settings.water_res_x*Settings.water_res_y*sin(lastWaterStepTime)*sin(lastWaterStepTime));

			rnd%=(Settings.water_res_x*Settings.water_res_y);

				//if not on border -> drop rain
			if(	rnd>Settings.water_res_x*2 &&
				rnd<(Settings.water_res_y-2)*Settings.water_res_x &&
				rnd%Settings.water_res_y>1 && rnd%Settings.water_res_y<Settings.water_res_y-2 )
				{
				w_current[rnd] = (int)(10000*sin(lastWaterStepTime*345)*10*sin(lastWaterStepTime/15));
				}
			}
		}


		//Ok now everything is updated, make changes visible...

		//calculate water height diffs
	w_current=water_data+(Settings.water_res_x*Settings.water_res_y)*((waterStep+1)&1);

	for(y=1;y<Settings.water_res_y-1;y++)
		{
		for(x=1;x<Settings.water_res_x-1;x++)
			{
			water_diff[y*Settings.water_res_x+x].u=(float)(w_current[y*Settings.water_res_x+x+1]-w_current[y*Settings.water_res_x+x])*0.00001f;
			water_diff[y*Settings.water_res_x+x].v=(float)(w_current[(y+1)*Settings.water_res_x+x]-w_current[y*Settings.water_res_x+x])*0.00001f;

			//water_diff[y*Settings.water_res_x+x].u=(float)50/100.0f;
			//water_diff[y*Settings.water_res_x+x].v=(float)-20/100.0f;
			}
		}


	tc_stepx=1.0f/((float)Settings.water_res_x-1);
	tc_stepy=1.0f/((float)Settings.water_res_y-1);

	//fill texcoords
	tc=water_tc;
	tc_diff=water_diff;

	for(y=0;y<Settings.water_res_y;y++)
		{
		tc_v=((float)y)*tc_stepy;
		//vtx_y=((float)y)*vtx_stepy+vtx_offsy;

		for(x=0;x<Settings.water_res_x;x++)
			{
			tc_u=((float)x)*tc_stepx;
			//vtx_x=((float)x)*vtx_stepx+vtx_offsx;

			tc->u=tc_u+tc_diff[0].u;tc->v=tc_v+tc_diff[0].v;tc++;
			//vtx->x=vtx_x;vtx->y=vtx_y;vtx++;

			tc_diff++;
			}
		}

	if(0 && OpenGLExtensions.ext_VBO) //FIXME: update texture coords
		{
		ext.glDeleteBuffersARB( 1, &vbo_tc );
		ext.glGenBuffersARB(1,&vbo_tc);
		ext.glBindBufferARB( GL_ARRAY_BUFFER_ARB, vbo_tc );
		ext.glBufferDataARB( GL_ARRAY_BUFFER_ARB, Settings.water_res_x*Settings.water_res_y*sizeof(TEXCOORD), water_tc, GL_DYNAMIC_DRAW_ARB);
		}

	return true;
	}

void GLWater::render()
	{
	//if(0)glClear( GL_COLOR_BUFFER_BIT ); //No need for clearing the screen as we draw over *everything* anyways

	glLoadIdentity();
	  glTranslatef( 0.0,0.0, -50.0f);

	glColor4f(1,1,1,1);

	glEnable(GL_TEXTURE_2D);
	glBindTexture( GL_TEXTURE_2D, Settings.bgTexID);

	if(OpenGLExtensions.ext_VBO) //use vertex buffers
		{
		glEnableClientState(GL_VERTEX_ARRAY);
		ext.glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo_vtx);
		glVertexPointer(2,GL_FLOAT,0,0);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		ext.glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo_tc);
		glTexCoordPointer(2,GL_FLOAT,0,0);

		glEnableClientState(GL_INDEX_ARRAY);
		ext.glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_idx);

        glDrawElements(GL_QUADS, (Settings.water_res_x-1)*(Settings.water_res_y-1)*4, GL_UNSIGNED_SHORT, 0);

		glDisableClientState(GL_INDEX_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	else	//send the vertex data to the graphics card
		{
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glVertexPointer(2,GL_FLOAT,0,water_vtx);
		glTexCoordPointer(2,GL_FLOAT,0,water_tc);

		glDrawElements(GL_QUADS,(Settings.water_res_x-1)*(Settings.water_res_y-1)*4,GL_UNSIGNED_SHORT,water_idx);

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}

	}





void GLWater::calcWaterStep() //simple, nice and relaxing water effect :)
	{
	WATERTYPE val=0;
	int x,y;

	WATERTYPE *w_new=water_data+(Settings.water_res_x*Settings.water_res_y)*(waterStep&1)    +Settings.water_res_x;
	WATERTYPE *w_old=water_data+(Settings.water_res_x*Settings.water_res_y)*((waterStep+1)&1) +Settings.water_res_x;

	for(y=1;y<Settings.water_res_y-1;y++)
		{
		for(x=0;x<Settings.water_res_x;x++)
			{
			// (old[-]+old[+]+old+old)/2-current[0]
			val=-w_new[0]+(w_old[-Settings.water_res_x]+w_old[-1]+w_old[Settings.water_res_x]+w_old[1])/2;
			*w_new=val-(val>>5);
							//TODO: ideas on how to only update parts of water array to make this even faster?
			w_new++;
			w_old++;
			}
		}

	waterStep++;
	}








