/***************************************************************************
 *   Copyright (C) 2004 by Stefan Baur                                     *
 *   stefan@c0n.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qapplication.h>
#include <qpixmap.h>

#include "kglwater.h"

KGLWater::KGLWater(QWidget *parent, const char *name)
		: QGLWidget(parent, name)
	{
	setWFlags(getWFlags()|WNoAutoErase); //FIXME: is this good? ... we want as few qt drawing as possible
	}


KGLWater::~KGLWater()
	{
	water.quit();
	}



void KGLWater::readSettings(KConfig* config)
	{
	QString imgFile=config->readPathEntry( "background", "desktop" );

	water.Settings.realtime=config->readBoolEntry("realtime",DEFAULT_GL_WATER_REALTIME);
	water.Settings.water_res_x=config->readNumEntry("water_res_x",DEFAULT_GL_WATER_RES_X);
	water.Settings.water_res_y=config->readNumEntry("water_res_y",DEFAULT_GL_WATER_RES_Y);
	water.Settings.water_step_time=1.0/(double)config->readNumEntry("water_cps",DEFAULT_GL_WATER_CPS);
	water.Settings.timer_ms=config->readNumEntry("timer_ms",DEFAULT_GL_WATER_TIMER_MS);

	water.Settings.effect_whirl=config->readBoolEntry("effect_whirl",DEFAULT_GL_WATER_EFFECT_WHIRL);
	water.Settings.effect_bigblob=config->readBoolEntry("effect_bigblob",DEFAULT_GL_WATER_EFFECT_BIGBLOB);
	water.Settings.effect_rain=config->readBoolEntry("effect_rain",DEFAULT_GL_WATER_EFFECT_RAIN);

	if(imgFile=="desktop")
		{
		bgImage=QPixmap::grabWindow(QApplication::desktop()->winId());
		}
	else
		{
		if(!bgImage.load( imgFile ))
			{
				//no such image ... grab desktop instead...
			bgImage=QPixmap::grabWindow(QApplication::desktop()->winId());
			}
		}

//	setBackgroundColor( black ); //FIXME: we no not need this... or?
//	erase();

	startTimer( water.Settings.timer_ms );
	}


void KGLWater::initializeGL()
	{
	int imgSize,maxTexture,maxDesktop,maxImage;

	water.init();

	if(bgImage.isNull()) //FIXME: whoo ... we do not have a background image!?
		{
		bgImage.create(64,64,32); //create a empty one - better ideas? (maybe fill some error text into?)
		bgImage.fill( 0xFF3311 );
		}

	glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxTexture);
	if(maxTexture<64)maxTexture=64; //we want at least 64x64 textures as it makes no sense to have less

	maxDesktop=QApplication::desktop()->width();
	if(QApplication::desktop()->height()>maxDesktop)maxDesktop=QApplication::desktop()->height();

	maxImage=bgImage.width();
	if(bgImage.height()>maxImage)maxImage=bgImage.height();

	qDebug(" maxTexture: %i, maxImage: %i, maxDesktop: %i",maxTexture,maxImage,maxDesktop);

		// increase imgSize by power of 2 to find the best fitting texture resolution for current image
	for(imgSize=64;imgSize<maxTexture;imgSize*=2)
		{
			//are we already as big as the image? Then break!
		if( imgSize>=maxImage )
			break;

			//are we already near the desktop resolution? Then break!
		if( imgSize*1.5>maxDesktop ) //*2 ?
			break;
		}

			// imgSize should now be a power of two - big enough for the image and desktop
	if(imgSize>maxTexture)imgSize=maxTexture;
	qDebug(" using %i x %i Texture!",imgSize,imgSize);

		//make it OpenGL compatible
	bgImage=bgImage.smoothScale(imgSize,imgSize,QImage::ScaleFree);
	bgImage=QGLWidget::convertToGLFormat(bgImage);

		//and load it into the graphics card
	glGenTextures(1,&water.Settings.bgTexID);
	glBindTexture(GL_TEXTURE_2D,water.Settings.bgTexID);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP); //GL_CLAMP or GL_REPEAT
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glEnable(GL_TEXTURE_2D);
	glTexImage2D(GL_TEXTURE_2D,0,3,bgImage.width(),bgImage.height(),0,GL_RGBA,GL_UNSIGNED_BYTE,bgImage.bits() );
	}


void KGLWater::resizeGL( int width, int height )
	{
		//set up for fine 2d rendering in the area [ -1/-1 , 1,1 ]

	glViewport( 0, 0, width, height );

	glMatrixMode(GL_PROJECTION);
	 glLoadIdentity();
	 glOrtho( -1, 1, -1, 1, 5.0, 60.0 );

	glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
	}


void KGLWater::paintGL()
	{
	water.render();
	}

void KGLWater::timerEvent(QTimerEvent*)
	{
	if(water.update())updateGL(); //TODO: dynamic timer delay
	}


#include "kglwater.moc"
